from typing import Any, Dict, NamedTuple, Optional

import numpy as np

from .compression import decode_data_blob


def _to_volts(value: Optional[float]):
    """Convert µV to V if not None."""
    if value is None:
        return None
    return float(value) / 1e6


def _to_seconds(value: Optional[float]):
    """Convert µs to s if not None."""
    if value is None:
        return None
    return float(value) / 1e6


class HitRecord(NamedTuple):
    """
    Hit record in pridb (SetType = 2).
    """

    time: float  #: Time in seconds
    channel: int  #: Channel number
    param_id: int  #: Parameter ID of table ae_params for ADC value conversion
    amplitude: float  #: Peak amplitude in volts
    duration: float  #: Hit duration in seconds
    energy: float  #: Energy (EN 1330-9) in eu (1e-14 V²s)
    rms: float  #: RMS of the noise before the hit in volts
    # optional for creating:
    set_id: Optional[int] = None  #: Unique identifier for data set in pridb
    threshold: Optional[float] = None  #: Threshold amplitude in volts
    rise_time: Optional[float] = None  #: Rise time in seconds
    signal_strength: Optional[float] = None  #: Signal strength in nVs (1e-9 Vs)
    counts: Optional[int] = None  #: Number of positive threshold crossings
    trai: Optional[int] = None  #: Transient recorder index (foreign key between pridb and tradb)
    cascade_hits: Optional[int] = None  #: Total number of hits in the same hit-cascade
    cascade_counts: Optional[int] = None  #: Summed counts of hits in the same hit-cascade
    cascade_energy: Optional[int] = None  #: Summed energy of hits in the same hit-cascade
    cascade_signal_strength: Optional[int] = None  #: Summed signal strength of hits in the same hit-cascade  # noqa  # pylint: disable=line-too-long

    @classmethod
    def from_sql(cls, row: Dict[str, Any]) -> "HitRecord":
        """
        Create HitRecord from SQL row.

        Args:
            row: Dict of column names and values
        """
        return cls(
            set_id=row["SetID"],
            time=row["Time"],
            channel=row["Chan"],
            param_id=row["ParamID"],
            threshold=_to_volts(row.get("Thr")),  # optional
            amplitude=_to_volts(row["Amp"]),
            rise_time=_to_seconds(row.get("RiseT")),  # optional
            duration=_to_seconds(row["Dur"]),
            energy=row["Eny"],
            signal_strength=row.get("SS"),  # optional for spotWave
            rms=_to_volts(row["RMS"]),
            counts=row.get("Counts"),  # optional
            trai=row.get("TRAI"),  # optional
            cascade_hits=row.get("CHits"),  # optional
            cascade_counts=row.get("CCnt"),  # optional
            cascade_energy=row.get("CEny"),  # optional
            cascade_signal_strength=row.get("CSS"),  # optional
        )


class MarkerRecord(NamedTuple):
    """
    Marker record in pridb (SetType = 4, 5, 6).

    A marker can have different meanings depending on its SetType:\n
    - 4: label\n
    - 5: datetime data set, as it is inserted whenever recording is started by software\n
    - 6: a section start marker. E.g. new sections are started, if acquisition settings changed
    """

    time: float  #: Time in seconds
    set_type: int  #: Marker type (see above)
    data: str  #: Content of marker (label text or datetime)
    # optional for creating:
    number: Optional[int] = None  #: Marker number
    set_id: Optional[int] = None  #: Unique identifier for data set in pridb

    @classmethod
    def from_sql(cls, row: Dict[str, Any]) -> "MarkerRecord":
        """
        Create MarkerRecord from SQL row.

        Args:
            row: Dict of column names and values
        """
        return cls(
            set_id=row["SetID"],
            time=row["Time"],
            set_type=row["SetType"],
            number=row["Number"],
            data=row["Data"],
        )


class StatusRecord(NamedTuple):
    """
    Status data record in pridb (SetType = 3).
    """

    time: float  #: Time in seconds
    channel: int  #: Channel number
    param_id: int  #: Parameter ID of table ae_params for ADC value conversion
    energy: float  #: Energy (EN 1330-9) in eu (1e-14 V²s)
    rms: float  #: RMS in volts
    # optional for creating:
    set_id: Optional[int] = None  #: Unique identifier for data set in pridb
    threshold: Optional[float] = None  #: Threshold amplitude in volts
    signal_strength: Optional[float] = None  #: Signal strength in nVs (1e-9 Vs)

    @classmethod
    def from_sql(cls, row: Dict[str, Any]) -> "StatusRecord":
        """
        Create StatusRecord from SQL row.

        Args:
            row: Dict of column names and values
        """
        return cls(
            set_id=row["SetID"],
            time=row["Time"],
            channel=row["Chan"],
            param_id=row["ParamID"],
            threshold=_to_volts(row.get("Thr")),  # optional
            energy=row["Eny"],
            signal_strength=row.get("SS"),  # optional for spotWave
            rms=_to_volts(row["RMS"]),  # optional
        )


class ParametricRecord(NamedTuple):
    """
    Parametric data record in pridb (SetType = 1).
    """

    time: float  #: Time in seconds
    param_id: int  #: Parameter ID of table ae_params for ADC value conversion
    # optional for creating:
    set_id: Optional[int] = None  #: Unique identifier for data set in pridb
    pctd: Optional[int] = None  #: Digital counter value
    pcta: Optional[int] = None  #: Analog hysteresis counter
    pa0: Optional[int] = None  #: Amplitude of parametric input 0 in volts
    pa1: Optional[int] = None  #: Amplitude of parametric input 1 in volts
    pa2: Optional[int] = None  #: Amplitude of parametric input 2 in volts
    pa3: Optional[int] = None  #: Amplitude of parametric input 3 in volts
    pa4: Optional[int] = None  #: Amplitude of parametric input 4 in volts
    pa5: Optional[int] = None  #: Amplitude of parametric input 5 in volts
    pa6: Optional[int] = None  #: Amplitude of parametric input 6 in volts
    pa7: Optional[int] = None  #: Amplitude of parametric input 7 in volts

    @classmethod
    def from_sql(cls, row: Dict[str, Any]) -> "ParametricRecord":
        """
        Create ParametricRecord from SQL row.

        Args:
            row: Dict of column names and values
        """
        return cls(
            set_id=row["SetID"],
            time=row["Time"],
            param_id=row["ParamID"],
            pctd=row.get("PCTD"),  # optional
            pcta=row.get("PCTA"),  # optional
            pa0=row.get("PA0"),  # optional
            pa1=row.get("PA1"),  # optional
            pa2=row.get("PA2"),  # optional
            pa3=row.get("PA3"),  # optional
            pa4=row.get("PA4"),  # optional
            pa5=row.get("PA5"),  # optional
            pa6=row.get("PA6"),  # optional
            pa7=row.get("PA7"),  # optional
        )


class TraRecord(NamedTuple):
    """
    Transient data record in tradb.

    Todo:
        Remove RMS
    """

    time: float  #: Time in seconds
    channel: int  #: Channel number
    param_id: int  #: Parameter ID of table tr_params for ADC value conversion
    pretrigger: int  #: Pretrigger samples
    threshold: float  #: Threshold amplitude in volts
    samplerate: int  #: Samplerate in Hz
    samples: int  #: Number of samples
    data_format: int  #: Data format (0 = uncompressed, 2 = FLAC compression)
    data: np.ndarray  #: Transient signal in volts
    # optional for creating:
    trai: Optional[int] = None  #: Transient recorder index (foreign key between pridb and tradb)
    rms: Optional[float] = None  #: RMS of the noise before the hit

    @classmethod
    def from_sql(cls, row: Dict[str, Any]) -> "TraRecord":
        return TraRecord(
            time=row["Time"],
            channel=row["Chan"],
            param_id=row["ParamID"],
            pretrigger=row["Pretrigger"],
            threshold=_to_volts(row["Thr"]),
            samplerate=row["SampleRate"],
            samples=row["Samples"],
            data_format=row["DataFormat"],
            data=decode_data_blob(row["Data"], row["DataFormat"], row["TR_mV"]),
            trai=row["TRAI"],
        )


class FeatureRecord(NamedTuple):
    """
    Transient feature record in trfdb.
    """

    trai: int  #: Transient recorder index
    features: Dict[str, float]  #: Feature dictionary (feature name -> value)

    @classmethod
    def from_sql(cls, row: Dict[str, Any]) -> "FeatureRecord":
        return FeatureRecord(
            trai=row.pop("TRAI"),
            features=row,
        )



from typing import Generic, Iterator, TypeVar

from typing_extensions import Protocol

T = TypeVar("T", covariant=True)
class SizedIterable(Protocol, Generic[T]):  # noqa: E302
    """Generic iterable, sized type that implements `__iter__` and `__len__`."""

    def __len__(self) -> int:  # pragma: no cover
        ...

    def __iter__(self) -> Iterator[T]:  # pragma: no cover
        ...



import sqlite3
from abc import ABCMeta, abstractmethod
from ast import literal_eval
from functools import wraps
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Set, Tuple, Union

from ._sql import ConnectionWrapper, insert_from_dict, read_sql_generator, update_from_dict


def require_write_access(func):
    @wraps(func)
    def wrapper(self: "Database", *args, **kwargs):
        if self._readonly:  # pylint: disable=protected-access
            raise ValueError(
                "Can not write to database in read-only mode. Open database with mode='rw'"
            )
        return func(self, *args, **kwargs)
    return wrapper


class Database:
    """Database base class for pridb, tradb and trfdb."""

    __metaclass__ = ABCMeta

    def __init__(
        self,
        filename: str,
        mode: str = "ro",
        *,
        table_prefix: str,
        required_file_ext: Optional[str] = None,
    ):
        # check file extension
        if required_file_ext is not None:
            file_ext = Path(filename).suffix
            if file_ext.lower() != required_file_ext.lower():
                raise ValueError(
                    f"File {filename} does not have the required extension {required_file_ext}"
                )

        if mode == "rwc":
            if not Path(filename).exists():
                self.create(filename)  # call abstract method (implemented by child class)

        self._readonly = (mode == "ro")
        self._connection_wrapper = ConnectionWrapper(filename, mode)

        self._table_prefix: str = table_prefix
        self._table_main: str = f"{table_prefix}_data"
        self._table_fieldinfo: str = f"{table_prefix}_fieldinfo"
        self._table_globalinfo: str = f"{table_prefix}_globalinfo"
        self._table_params: str = f"{table_prefix}_params"

        # check if required tables exist
        for table in (self._table_main, self._table_fieldinfo, self._table_globalinfo):
            if table not in self.tables():
                raise ValueError(f"Required table {table} not found in database")

        # cached results
        self._parameter_table_cached: Dict[int, Dict[str, Any]] = {}

    @staticmethod
    @abstractmethod
    def create(filename: str):
        """
        Create empty database.

        Static method has to be implemented by child classes.
        """
        raise NotImplementedError("Database creation not implemented")

    @property
    def filename(self) -> str:
        """Filename of database."""
        return self._connection_wrapper.filename

    @property
    def connected(self) -> bool:
        """Check if connected to SQLite database."""
        return self._connection_wrapper.connected

    def connection(self) -> sqlite3.Connection:
        """
        Get SQLite connection object.

        Raises:
            RuntimeError: If connection is closed
        """
        return self._connection_wrapper.connection()

    def rows(self) -> int:
        """Number of rows in data table."""
        con = self.connection()
        cur = con.execute(f"SELECT COUNT(*) FROM {self._table_main}")
        return cur.fetchone()[0]

    def _columns(self, table: str) -> Tuple[str, ...]:
        """Columns of specified table."""
        con = self.connection()
        cur = con.execute(f"SELECT * FROM {table} LIMIT 0")  # empty dummy query
        return tuple(str(column[0]) for column in cur.description)

    def columns(self) -> Tuple[str, ...]:
        """Columns of data table."""
        return self._columns(self._table_main)

    @require_write_access
    def _add_columns(
        self,
        table: str,
        columns: Union[str, Sequence[str]],
        dtype: Optional[str] = None,
    ):
        """Add columns to specified table."""
        if dtype is None:
            dtype = ""
        with self.connection() as con:  # commit/rollback transaction
            columns_exist = self._columns(table)
            for column in columns:  # keep order of columns
                if column not in columns_exist:
                    con.execute(f"ALTER TABLE {table} ADD COLUMN {column} {dtype}")

    def tables(self) -> Set[str]:
        """Get table names."""
        con = self.connection()
        cur = con.execute("SELECT name FROM sqlite_master WHERE type == 'table'")
        tables = {result[0] for result in cur.fetchall()}
        return tables

    def fieldinfo(self) -> Dict[str, Dict[str, Any]]:
        """
        Read fieldinfo table.

        The fieldinfo table stores informations about columns of the data table (like units).

        Returns:
            Dict of column names and informations (again a dict)
        """
        con = self.connection()
        query = f"SELECT * FROM {self._table_fieldinfo}"
        return {row.pop("field"): row for row in read_sql_generator(con, query)}

    @require_write_access
    def write_fieldinfo(self, field: str, info: Dict[str, Any]):
        """
        Write to fieldinfo table.

        Args:
            field: Column name of data table
            info: Dict of properties and values, e.g. {"Unit": "[Hz]"}

        Raises:
            ValueError: If field is not a column of data table
        """
        if field not in self.columns():
            raise ValueError(f"Field {field} must be a column of data table")

        row_dict = info
        row_dict["field"] = field
        with self.connection() as con:  # commit/rollback transaction
            try:
                if field in self.fieldinfo().keys():
                    update_from_dict(con, self._table_fieldinfo, row_dict, "field")
                else:
                    insert_from_dict(con, self._table_fieldinfo, row_dict)
            except sqlite3.OperationalError:  # missing column(s)
                self._add_columns(self._table_fieldinfo, list(row_dict.keys()))
                self.write_fieldinfo(field, info)  # try again

    def globalinfo(self) -> Dict[str, Any]:
        """Read globalinfo table."""
        def try_convert_string(value: str) -> Any:
            try:
                return literal_eval(value)
            except (SyntaxError, ValueError):
                return str(value)
        con = self.connection()
        cur = con.execute(f"SELECT Key, Value FROM {self._table_globalinfo}")
        return {
            row[0]: try_convert_string(str(row[1])) for row in cur.fetchall()
        }

    @require_write_access
    def _update_globalinfo(self):
        """Update globalinfo after writes."""
        keys = self.globalinfo().keys()
        with self.connection() as con:  # commit/rollback transaction
            if "ValidSets" in keys:
                con.execute(
                    """
                    UPDATE {prefix}_globalinfo
                    SET Value = (SELECT MAX(rowid) FROM {prefix}_data)
                    WHERE Key == "ValidSets"
                    """.format(prefix=self._table_prefix)
                )
            if "TRAI" in keys:
                con.execute(
                    """
                    UPDATE {prefix}_globalinfo
                    SET Value = (SELECT MAX(TRAI) FROM {prefix}_data)
                    WHERE Key == "TRAI";
                    """.format(prefix=self._table_prefix)
                )

    def _file_status(self) -> int:
        """Get file status (0: offline, 1: suspended, 2: active)."""
        result = self.connection().execute(
            f"SELECT Value FROM {self._table_globalinfo} WHERE Key == 'FileStatus'"
        ).fetchone()
        return int(result[0]) if result else 0

    def _main_index_range(self) -> Tuple[int, int]:
        """Get range of main data table index (SetID for pridb/tradb or TRAI for trfdb)."""
        return self.connection().execute(
            f"SELECT MIN(rowid), MAX(rowid) FROM {self._table_main}"
        ).fetchone()

    def _parameter_table(self) -> Dict[int, Dict[str, Any]]:
        """Read *_params table to dict."""
        def parameter_by_id():
            for row in read_sql_generator(
                self.connection(),
                f"SELECT * FROM {self._table_params}",
            ):
                param_id = row.pop("ID")
                yield param_id, row

        if not self._parameter_table_cached:
            self._parameter_table_cached = dict(parameter_by_id())
        return self._parameter_table_cached

    def _parameter(self, param_id: int) -> Dict[str, Any]:
        """Read parameters from *_params by ID."""
        try:
            return self._parameter_table()[param_id]
        except KeyError:
            raise ValueError(
                f"Parameter ID {param_id} not found in {self._table_params}"
            ) from None

    def close(self):
        """Close database connection."""
        if not hasattr(self, "_connection_wrapper"):
            return
        if self.connected:
            if not self._readonly:
                self._update_globalinfo()
            self._connection_wrapper.close()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()



from functools import lru_cache
from itertools import chain
from pathlib import Path
from time import sleep
from typing import Iterable, Optional, Sequence, Set, Tuple, Union

import numpy as np
import pandas as pd
from tqdm import tqdm

from ._database import Database, require_write_access
from ._dataframe import iter_to_dataframe
from ._sql import (
    QueryIterable,
    create_new_database,
    insert_from_dict,
    query_conditions,
    read_sql_generator,
    sql_binary_search,
)
from .compression import encode_data_blob
from .datatypes import TraRecord
from .types import SizedIterable


@lru_cache(maxsize=32, typed=True)
def _create_time_vector(
    samples: int, samplerate: int, pretrigger: int = 0
) -> np.ndarray:
    t_raw = np.arange(-pretrigger, samples - pretrigger, dtype=np.int32)
    t_scaled = (t_raw / np.float32(samplerate)).astype(np.float32)
    return t_scaled


class TraDatabase(Database):
    """IO Wrapper for tradb database file."""

    def __init__(
        self, filename: str, mode: str = "ro", *, compression: bool = False,
    ):
        """
        Open tradb database file.

        Args:
            filename: Path to tradb database file
            mode: Define database access:
                **"ro"** (read-only),
                **"rw"** (read-write),
                **"rwc"** (read-write and create empty database if it does not exist)
            compression: Enable/disable FLAC compression data BLOBs for writing
        """
        super().__init__(
            filename, mode=mode, table_prefix="tr", required_file_ext=".tradb",
        )
        self._data_format = 2 if compression else 0
        self._timebase = self.globalinfo()["TimeBase"]

    @staticmethod
    def create(filename: str):
        """
        Create empty tradb.

        Args:
            filename: Path to new tradb database file
        """
        file_schema = Path(__file__).resolve().parent / "schema_templates/tradb.sql"
        with open(file_schema, "r", encoding="utf-8") as file:
            schema_tradb = file.read()
        schema_tradb = schema_tradb.format(timebase=int(1e7))  # fill placeholder / constants
        create_new_database(filename, schema_tradb)

    def channel(self) -> Set[int]:
        """Get list of channels."""
        con = self.connection()
        cur = con.execute("SELECT DISTINCT Chan FROM tr_data WHERE Chan IS NOT NULL")
        return {result[0] for result in cur.fetchall()}

    def read(self, **kwargs) -> pd.DataFrame:
        """
        Read transient data to Pandas DataFrame.

        Args:
            **kwargs: Arguments passed to `iread`

        Returns:
            Pandas DataFrame with transient data
        """
        return iter_to_dataframe(
            self.iread(**kwargs), desc="Tra", index_column="trai",
        )

    def _get_total_time_range(self) -> Tuple[float, float]:
        """Return total time range [min, max] of tradb."""
        def get_time(func: str):
            result = self.connection().execute(
                f"SELECT Time FROM tr_data WHERE TRAI == (SELECT {func}(TRAI) from tr_data)"
            ).fetchone()
            return 0 if result is None else result[0] / self._timebase
        return get_time("MIN"), get_time("MAX")

    def _get_trai_range_from_time_range(
        self, time_start: Optional[float], time_stop: Optional[float]
    ) -> Tuple[Optional[int], Optional[int]]:
        """Use binary search to find indexes (TRAI) of a given time range."""
        con = self.connection()
        trai_start = None
        trai_stop = None
        if time_start is not None:
            trai_start = sql_binary_search(
                connection=con,
                table="tr_data",
                column_value="Time",
                column_index="TRAI",
                fun_compare=lambda t: t >= time_start * self._timebase,  # type: ignore
                lower_bound=True,  # return lower index of true conditions
            )
        if time_stop is not None:
            trai_stop = sql_binary_search(
                connection=con,
                table="tr_data",
                column_value="Time",
                column_index="TRAI",
                fun_compare=lambda t: t < time_stop * self._timebase,  # type: ignore
                lower_bound=False,  # return upper index of true conditions
            )
        return trai_start, trai_stop

    def iread(
        self,
        *,
        channel: Union[None, int, Sequence[int]] = None,
        time_start: Optional[float] = None,
        time_stop: Optional[float] = None,
        trai: Union[None, int, Sequence[int]] = None,
        query_filter: Optional[str] = None,
    ) -> SizedIterable[TraRecord]:
        """
        Stream transient data with returned Iterable.

        Args:
            channel: None if all channels should be read.
                Otherwise specify the channel number or a list of channel numbers
            time_start: Start reading at relative time (in seconds). Start at beginning if `None`
            time_stop: Stop reading at relative time (in seconds). Read until end if `None`
            trai: Read data by TRAI (transient recorder index)
            query_filter: Optional query filter provided as SQL clause,
                e.g. "Pretrigger == 500 AND Samples >= 1024"

        Returns:
            Sized iterable to sequential read transient data
        """
        # check for empty time ranges
        time_min, time_max = self._get_total_time_range()
        if time_start is not None:
            if time_start > time_max:
                return []
        if time_stop is not None:
            if time_stop < time_min:
                return []
        if time_start is not None and time_stop is not None:
            if time_start >= time_stop:
                return []

        trai_start, trai_stop = self._get_trai_range_from_time_range(time_start, time_stop)
        # nested query to fix ambiguous column name error with query_filter
        query = """
        SELECT * FROM (
            SELECT vtr.*, tr.ParamID
            FROM view_tr_data vtr
            LEFT JOIN tr_data tr ON vtr.SetID == tr.SetID
            ORDER BY TRAI ASC
        )
        """ + query_conditions(
            isin={"Chan": channel, "TRAI": trai},
            greater_equal={"TRAI": trai_start},
            # < condition already met in binary search, use <= here for found indice range
            less_equal={"TRAI": trai_stop},
            custom_filter=query_filter,
        )
        return QueryIterable(
            self._connection_wrapper.get_readonly_connection(),
            query,
            TraRecord.from_sql,
        )

    def read_wave(
        self, trai: int, time_axis: bool = True,
    ) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, int]]:
        """
        Read transient signal for a given TRAI (transient recorder index).

        This method is useful in combination with `PriDatabase.read_hits`,
        that will store the TRAI in a DataFrame.

        Args:
            trai: Transient recorder index (unique key between pridb and tradb)
            time_axis: Create the correspondig time axis. Default: `True`

        Returns:
            If :attr:`time_axis` is `True`\n
            - Array with transient signal
            - Time axis

            If :attr:`time_axis` is `False`\n
            - Array with transient signal
            - Samplerate
        """
        iterable = self.iread(trai=trai)
        try:
            tra = next(iter(iterable))
        except StopIteration:
            raise ValueError("TRAI does not exists") from None

        if time_axis:
            return (
                tra.data,
                _create_time_vector(tra.samples, tra.samplerate, tra.pretrigger),
            )
        return tra.data, tra.samplerate

    def _get_previous_trai(self, channel: int, trai: int) -> Optional[int]:
        """Find previous tra record index for given channel and TRAI."""
        result = self.connection().execute(
            "SELECT TRAI FROM tr_data WHERE Chan == ? AND TRAI < ? ORDER BY TRAI DESC LIMIT 1",
            (channel, trai),
        ).fetchone()
        return result[0] if result is not None else None

    def read_continuous_wave(  # pylint: disable=too-many-locals
        self,
        channel: int,
        time_start: Optional[float] = None,
        time_stop: Optional[float] = None,
        *,
        time_axis: bool = True,
        show_progress: bool = True,
    ) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, int]]:
        """
        Read transient signal of specified channel to a single, continuous array.

        The signal is exactly cropped to the given time range. Time gaps are filled with 0's.

        Args:
            channel: Channel number to read
            time_start: Start reading at relative time (in seconds). Start at beginning if `None`
            time_stop: Stop reading at relative time (in seconds). Read until end if `None`
            time_axis: Create the correspondig time axis. Default: `True`
            show_progress: Show progress bar. Default: `True`

        Returns:
            If `time_axis` is `True`\n
            - Array with transient signal
            - Time axis

            If `time_axis` is `False`\n
            - Array with transient signal
            - Samplerate
        """
        iterable = self.iread(channel=channel, time_start=time_start, time_stop=time_stop)
        iterator = iter(iterable)
        if show_progress:
            iterator = tqdm(iterator, total=len(iterable), desc="Tra")  # ignores previous tra

        # prepend previous tra to iterator if available
        trai_start, _ = self._get_trai_range_from_time_range(time_start, None)
        if trai_start is not None:
            previous_trai = self._get_previous_trai(channel, trai_start)
            if previous_trai is not None:
                iterator = chain(
                    iter(self.iread(channel=channel, trai=previous_trai)),
                    iterator,
                )

        # find beginning if not provided
        if time_start is None:
            time_start = self._get_total_time_range()[0]

        def slice_range(tra: TraRecord):
            """Get indices to slice given tra record data to time range."""
            def limit_index(i: int):
                return min(max(0, i), tra.samples)
            n_start, n_stop = 0, tra.samples
            if time_start is not None:
                n_start = limit_index(round((time_start - tra.time) * tra.samplerate))
            if time_stop is not None:
                n_stop = limit_index(round((time_stop - tra.time) * tra.samplerate))
            return n_start, n_stop

        samplerate = 0  # will be initialized with samplerate of first record
        tra_blocks = [np.empty(0, dtype=np.float32)]
        expected_time = time_start

        for tra in iterator:
            if samplerate == 0:
                samplerate = tra.samplerate
            if tra.samplerate != samplerate:
                raise RuntimeError("Different sampling rates inside requested time interval")

            # check for gaps in tra stream
            time_gap = tra.time - expected_time
            if time_gap > 1 / tra.samplerate:
                samples_gap = round(time_gap * tra.samplerate)
                tra_blocks.append(np.zeros(samples_gap, dtype=np.float32))

            sample_start, sample_stop = slice_range(tra)
            tra_blocks.append(tra.data[sample_start:sample_stop])
            expected_time = tra.time + sample_stop / tra.samplerate
            if time_start is not None:
                # the prepended record might be out of time range -> avoid exceeding zero-padding
                expected_time = max(expected_time, time_start)

        if time_stop is not None and samplerate and abs(expected_time - time_stop) > 1 / samplerate:
            # zero padding at ending
            samples = round((time_stop - expected_time) * samplerate)
            tra_blocks.append(np.zeros(samples, dtype=np.float32))

        y = np.concatenate(tra_blocks)
        if time_axis:
            return y, _create_time_vector(len(y), samplerate) + time_start
        return y, samplerate

    def listen(
        self,
        existing: bool = False,
        wait: bool = False,
        query_filter: Optional[str] = None,
    ) -> Iterable[TraRecord]:
        """
        Listen to database changes and return new records.

        Args:
            existing: Return already existing records
            wait: Wait for new records even if no acquisition (writer) is active.
                Otherwise the function returns after all records are read.
            query_filter: Optional query filter provided as SQL clause,
                e.g. "TRAI >= 100 AND Samples >= 1024"

        Yields:
            New transient data records
        """
        query = """
        SELECT * FROM (
            SELECT vtr.*, tr.ParamID
            FROM view_tr_data vtr
            LEFT JOIN tr_data tr ON vtr.SetID == tr.SetID
            WHERE vtr.SetID > ?
        )
        """ + query_conditions(custom_filter=query_filter)
        last_set_id = 0 if existing else self._main_index_range()[1]
        while True:
            file_status = self._file_status()
            for row in list(read_sql_generator(self.connection(), query, last_set_id)):
                yield TraRecord.from_sql(row)
                last_set_id = row["SetID"]
            if not wait and file_status == 0:  # no writer active
                break
            sleep(0.1)  # wait 100 ms until next read


    @require_write_access
    def write(self, tra: TraRecord) -> int:
        """
        Write transient data to pridb.

        Args:
            tra: Transient data set

        Returns:
            Index (SetID) of inserted row

        Todo:
            Status flag
        """
        # self._validate_and_update_time(tra.time)
        parameter = self._parameter(tra.param_id)
        with self.connection() as con:  # commit/rollback transaction
            return insert_from_dict(
                con,
                self._table_main,
                {
                    "Time": int(tra.time * self._timebase),
                    "Chan": int(tra.channel),
                    "Status": 32768,
                    "ParamID": int(tra.param_id),
                    "Pretrigger": int(tra.pretrigger),
                    "Thr": int(tra.threshold * 1e6 / parameter["ADC_µV"]),
                    "SampleRate": int(tra.samplerate),
                    "Samples": int(tra.samples),
                    "DataFormat": int(self._data_format),
                    "Data": encode_data_blob(tra.data, self._data_format, parameter["TR_mV"]),
                    "TRAI": (
                        int(tra.trai)
                        if tra.trai else None
                    ),
                },
            )



from functools import wraps
from pathlib import Path
from time import sleep
from typing import Iterable, Optional, Sequence, Set, Union

import pandas as pd

from ._database import Database, require_write_access
from ._dataframe import iter_to_dataframe
from ._sql import (
    QueryIterable,
    create_new_database,
    insert_from_dict,
    query_conditions,
    read_sql_generator,
)
from .datatypes import HitRecord, MarkerRecord, ParametricRecord, StatusRecord
from .types import SizedIterable

RecordType = Union[HitRecord, MarkerRecord, ParametricRecord, StatusRecord]


def check_monotonic_time(func):
    def get_max_time(database: "PriDatabase"):
        con = database.connection()
        cur = con.execute("SELECT Time FROM view_ae_data ORDER BY SetID DESC LIMIT 1")
        try:
            return cur.fetchone()[0]  # None is not subscriptable -> TypeError
        except TypeError:
            return 0

    @wraps(func)
    def wrapper(self: "PriDatabase", record: RecordType, *args, **kwargs):
        max_time = get_max_time(self)
        if record.time + 1e-9 < max_time:  # threshold of 1 ns to ignore rounding errors
            raise ValueError(
                (
                    f"Time column has to be monotonic increasing. "
                    f"Time of current / last row: {record.time} / {max_time} s"
                )
            )
        return func(self, record, *args, **kwargs)
    return wrapper


class PriDatabase(Database):
    """IO Wrapper for pridb database file."""

    def __init__(self, filename: str, mode: str = "ro"):
        """
        Open pridb database file.

        Args:
            filename: Path to pridb database file
            mode: Define database access:
                **"ro"** (read-only),
                **"rw"** (read-write),
                **"rwc"** (read-write and create empty database if it does not exist)
        """
        super().__init__(
            filename, mode=mode, table_prefix="ae", required_file_ext=".pridb",
        )
        self._timebase = self.globalinfo()["TimeBase"]

    @staticmethod
    def create(filename: str):
        """
        Create empty pridb.

        Args:
            filename: Path to new pridb database file
        """
        file_schema = Path(__file__).resolve().parent / "schema_templates/pridb.sql"
        with open(file_schema, "r", encoding="utf-8") as file:
            schema_pridb = file.read()
        schema_pridb = schema_pridb.format(timebase=int(1e7))  # fill placeholder / constants
        create_new_database(filename, schema_pridb)

    def channel(self) -> Set[int]:
        """Get list of channels."""
        con = self.connection()
        cur = con.execute("SELECT DISTINCT Chan FROM ae_data WHERE Chan IS NOT NULL")
        return {result[0] for result in cur.fetchall()}

    def read_hits(self, **kwargs) -> pd.DataFrame:
        """
        Read hits to Pandas DataFrame.

        Args:
            **kwargs: Arguments passed to `iread_hits`

        Returns:
            Pandas DataFrame with hit data
        """
        return iter_to_dataframe(
            self.iread_hits(**kwargs), desc="Hits", index_column="set_id",
        )

    def read_markers(self, **kwargs) -> pd.DataFrame:
        """
        Read marker to Pandas DataFrame.

        Args:
            **kwargs: Arguments passed to `iread_markers`

        Returns:
            Pandas DataFrame with marker data
        """
        dtypes = {
            "number": pd.Int64Dtype(),
            "data": str,
        }
        return (
            iter_to_dataframe(
                self.iread_markers(**kwargs), desc="Marker", index_column="set_id",
            )
            .apply(lambda x: x.astype(dtypes.get(x.name, x.dtype)))
        )

    def read_parametric(self, **kwargs) -> pd.DataFrame:
        """
        Read parametric data to Pandas DataFrame.

        Args:
            **kwargs: Arguments passed to `iread_parametric`

        Returns:
            Pandas DataFrame with parametric data
        """
        return iter_to_dataframe(
            self.iread_parametric(**kwargs), desc="Parametric", index_column="set_id",
        )

    def read_status(self, **kwargs) -> pd.DataFrame:
        """
        Read status data to Pandas DataFrame.

        Args:
            **kwargs: Arguments passed to `iread_status`

        Returns:
            Pandas DataFrame with status data
        """
        return iter_to_dataframe(
            self.iread_status(**kwargs), desc="Status", index_column="set_id",
        )

    def read(self, **kwargs) -> pd.DataFrame:
        """
        Read all data set types (hits, markers, parametric data, status data)
        from pridb to Pandas DataFrame.

        Args:
            **kwargs: Arguments passed to `iread_hits`, `iread_markers`, `iread_parametric`
                and `iread_status`

        Returns:
            Pandas DataFrame with all pridb data set types
        """
        # read to separate dataframes
        df_hits = self.read_hits(**kwargs)
        df_markers = self.read_markers(**kwargs)
        df_parametric = self.read_parametric(**kwargs)
        df_status = self.read_status(**kwargs)

        # add missing set_types
        column_set_type = 0
        if not df_hits.empty:
            df_hits.insert(column_set_type, "set_type", 2)
        if not df_parametric.empty:
            df_parametric.insert(column_set_type, "set_type", 1)
        if not df_status.empty:
            df_status.insert(column_set_type, "set_type", 3)

        # drop additional marker columns
        df_markers.drop(columns=["number", "data"], inplace=True, errors="ignore")

        # dict to restore original data types
        dtypes = {
            **dict(df_hits.dtypes),
            **dict(df_markers.dtypes),
            **dict(df_parametric.dtypes),
            **dict(df_status.dtypes),
        }

        # concat all dataframes, restore types and sort by index
        df = (
            pd.concat(
                [df_markers, df_hits, df_status, df_parametric], sort=False, copy=False
            )
            .apply(lambda x: x.astype(dtypes[x.name]))
            .sort_index()
            .rename_axis(df_hits.index.name)
        )
        return df

    def iread_hits(
        self,
        *,
        channel: Union[None, int, Sequence[int]] = None,
        time_start: Optional[float] = None,
        time_stop: Optional[float] = None,
        set_id: Union[None, int, Sequence[int]] = None,
        query_filter: Optional[str] = None,
    ) -> SizedIterable[HitRecord]:
        """
        Stream hits with returned iterable.

        Args:
            channel: None if all channels should be read.
                Otherwise specify the channel number or a list of channel numbers
            time_start: Start reading at relative time (in seconds). Start at beginning if `None`
            time_stop: Stop reading at relative time (in seconds). Read until end if `None`
            set_id: Read by SetID
            query_filter: Optional query filter provided as SQL clause,
                e.g. "Amp > 5000 AND RiseT < 1000"

        Returns:
            Sized iterable to sequential read hits
        """
        # nested query to fix ambiguous column name error with query_filter
        query = """
        SELECT * FROM (
            SELECT vae.*, ae.ParamID
            FROM view_ae_data vae
            LEFT JOIN ae_data ae ON vae.SetID == ae.SetID
        )
        """ + query_conditions(
            equal={"SetType": 2},
            isin={"Chan": channel, "SetID": set_id},
            greater_equal={"Time": time_start},
            less={"Time": time_stop},
            custom_filter=query_filter,
        )
        return QueryIterable(
            self._connection_wrapper.get_readonly_connection(),
            query,
            HitRecord.from_sql,
        )

    def iread_markers(
        self,
        *,
        time_start: Optional[float] = None,
        time_stop: Optional[float] = None,
        set_id: Union[None, int, Sequence[int]] = None,
        query_filter: Optional[str] = None,
    ) -> SizedIterable[MarkerRecord]:
        """
        Stream markers with returned iterable.

        Args:
            time_start: Start reading at relative time (in seconds). Start at beginning if `None`
            time_stop: Stop reading at relative time (in seconds). Read until end if `None`
            set_id: Read by SetID
            query_filter: Optional query filter provided as SQL clause,
                e.g. "Number > 11 AND Data LIKE '%TimeZone%'"

        Returns:
            Sized iterable to sequential read markers
        """
        query = """
        SELECT SetID, Time, SetType, Number, Data
        FROM view_ae_markers vae
        """ + query_conditions(
            isin={"SetID": set_id},
            greater_equal={"Time": time_start},
            less={"Time": time_stop},
            custom_filter=query_filter,
        )
        return QueryIterable(
            self._connection_wrapper.get_readonly_connection(),
            query,
            MarkerRecord.from_sql,
        )

    def iread_parametric(
        self,
        *,
        time_start: Optional[float] = None,
        time_stop: Optional[float] = None,
        set_id: Union[None, int, Sequence[int]] = None,
        query_filter: Optional[str] = None,
    ) -> SizedIterable[ParametricRecord]:
        """
        Stream parametric data with returned iterable.

        Args:
            time_start: Start reading at relative time (in seconds). Start at beginning if `None`
            time_stop: Stop reading at relative time (in seconds). Read until end if `None`
            set_id: Read by SetID
            query_filter: Optional query filter provided as SQL clause,
                e.g. "PA0 >= -5000 AND PA0 < 5000"

        Returns:
            Sized iterable to sequential read parametric data
        """
        # nested query to fix ambiguous column name error with query_filter
        query = """
        SELECT * FROM (
            SELECT vae.*, ae.ParamID
            FROM view_ae_data vae
            LEFT JOIN ae_data ae ON vae.SetID == ae.SetID
        )
        """ + query_conditions(
            equal={"SetType": 1},
            isin={"SetID": set_id},
            greater_equal={"Time": time_start},
            less={"Time": time_stop},
            custom_filter=query_filter,
        )
        return QueryIterable(
            self._connection_wrapper.get_readonly_connection(),
            query,
            ParametricRecord.from_sql,
        )

    def iread_status(
        self,
        *,
        channel: Union[None, int, Sequence[int]] = None,
        time_start: Optional[float] = None,
        time_stop: Optional[float] = None,
        set_id: Union[None, int, Sequence[int]] = None,
        query_filter: Optional[str] = None,
    ) -> SizedIterable[StatusRecord]:
        """
        Stream status data with returned iterable.

        Args:
            channel: None if all channels should be read.
                Otherwise specify the channel number or a list of channel numbers
            time_start: Start reading at relative time (in seconds). Start at beginning if `None`
            time_stop: Stop reading at relative time (in seconds). Read until end if `None`
            set_id: Read by SetID
            query_filter: Optional query filter provided as SQL clause,
                e.g. "RMS < 300 OR RMS > 500"

        Returns:
            Sized iterable to sequential read status data
        """
        # nested query to fix ambiguous column name error with query_filter
        query = """
        SELECT * FROM (
            SELECT vae.*, ae.ParamID
            FROM view_ae_data vae
            LEFT JOIN ae_data ae ON vae.SetID == ae.SetID
        )
        """ + query_conditions(
            equal={"SetType": 3},
            isin={"Chan": channel, "SetID": set_id},
            greater_equal={"Time": time_start},
            less={"Time": time_stop},
            custom_filter=query_filter,
        )
        return QueryIterable(
            self._connection_wrapper.get_readonly_connection(),
            query,
            StatusRecord.from_sql,
        )

    def listen(
        self,
        existing: bool = False,
        wait: bool = False,
        query_filter: Optional[str] = None,
    ) -> Iterable[Union[HitRecord, MarkerRecord, ParametricRecord, StatusRecord]]:
        """
        Listen to database changes and return new records.

        Args:
            existing: Return already existing records
            wait: Wait for new records even if no acquisition (writer) is active.
                Otherwise the function returns after all records are read.
            query_filter: Optional query filter provided as SQL clause,
                e.g. "Time >= 100 AND Chan == 2"

        Yields:
            New hit/marker/parametric/status data records
        """
        query = """
        SELECT * FROM (
            SELECT vae.*, ae.ParamID
            FROM view_ae_data vae
            LEFT JOIN ae_data ae ON vae.SetID == ae.SetID
            WHERE vae.SetID > ?
        )
        """ + query_conditions(custom_filter=query_filter)
        last_set_id = 0 if existing else self._main_index_range()[1]
        while True:
            file_status = self._file_status()
            for row in list(read_sql_generator(self.connection(), query, last_set_id)):
                if row["SetType"] == 1:
                    yield ParametricRecord.from_sql(row)
                elif row["SetType"] == 2:
                    yield HitRecord.from_sql(row)
                elif row["SetType"] == 3:
                    yield StatusRecord.from_sql(row)
                elif row["SetType"] in (4, 5, 6):
                    yield from self.iread_markers(set_id=row["SetID"])
                last_set_id = row["SetID"]
            if not wait and file_status == 0:  # no writer active
                break
            sleep(0.1)  # wait 100 ms until next read

    @require_write_access
    @check_monotonic_time
    def write_hit(self, hit: HitRecord):
        """
        Write hit to pridb.

        Caution: `HitRecord.set_id` is ignored and automatically incremented.

        Args:
            hit: Hit data set

        Returns:
            Index (SetID) of inserted row

        Todo:
            Status flag
        """
        parameter = self._parameter(hit.param_id)
        with self.connection() as con:  # commit/rollback transaction
            return insert_from_dict(
                con,
                self._table_main,
                {
                    "SetType": 2,
                    "Time": int(hit.time * self._timebase),
                    "Chan": int(hit.channel),
                    "Status": 0,
                    "ParamID": int(hit.param_id),
                    "Thr": (
                        int(hit.threshold * 1e6 / parameter["ADC_µV"])
                        if hit.threshold else None
                    ),
                    "Amp": int(hit.amplitude * 1e6 / parameter["ADC_µV"]),
                    "RiseT": (
                        int(hit.rise_time * self._timebase)
                        if hit.rise_time else None
                    ),
                    "Dur": int(hit.duration * self._timebase),
                    "Eny": int(hit.energy / parameter["ADC_TE"]),
                    "SS": int(hit.signal_strength / parameter["ADC_SS"]),
                    "RMS": int(hit.rms * 1e6 / parameter["ADC_µV"] / 0.0065536),
                    "Counts": (
                        int(hit.counts)
                        if hit.counts else None
                    ),
                    "TRAI": (
                        int(hit.trai)
                        if hit.trai else None
                    ),
                },
            )

    @require_write_access
    @check_monotonic_time
    def write_marker(self, marker: MarkerRecord):
        """
        Write marker to pridb.

        Caution: `MarkerRecord.set_id` is ignored and automatically incremented.

        Args:
            marker: Marker data set

        Returns:
            Index (SetID) of inserted row
        """
        with self.connection() as con:  # commit/rollback transaction
            set_id = insert_from_dict(
                con,
                self._table_main,
                {
                    "SetType": int(marker.set_type),
                    "Time": int(marker.time * self._timebase),
                },
            )
            insert_from_dict(
                con,
                "ae_markers",
                {
                    "SetID": int(set_id),
                    "Number": (
                        int(marker.number)
                        if marker.number else None
                    ),
                    "Data": marker.data,
                },
            )
            return set_id

    @require_write_access
    @check_monotonic_time
    def write_status(self, status: StatusRecord):
        """
        Write status data to pridb.

        Caution: `StatusRecord.set_id` is ignored and automatically incremented.

        Args:
            status: Status data set

        Returns:
            Index (SetID) of inserted row

        Todo:
            Status flag
        """
        parameter = self._parameter(status.param_id)
        with self.connection() as con:  # commit/rollback transaction
            return insert_from_dict(
                con,
                self._table_main,
                {
                    "SetType": 3,
                    "Time": int(status.time * self._timebase),
                    "Chan": int(status.channel),
                    "Status": 0,
                    "ParamID": int(status.param_id),
                    "Thr": (
                        int(status.threshold * 1e6 / parameter["ADC_µV"])
                        if status.threshold else None
                    ),
                    "Eny": int(status.energy / parameter["ADC_TE"]),
                    "SS": int(status.signal_strength / parameter["ADC_SS"]),
                    "RMS": int(status.rms * 1e6 / parameter["ADC_µV"] / 0.0065536),
                },
            )

    @require_write_access
    @check_monotonic_time
    def write_parametric(self, parametric: ParametricRecord):
        """
        Write parametric data to pridb.

        Caution: `ParametricRecord.set_id` is ignored and automatically incremented.

        Args:
            parametric: Parametric data set

        Returns:
            Index (SetID) of inserted row

        Todo:
            Status flag
        """
        parameter = self._parameter(parametric.param_id)
        def try_convert(value: Optional[int], conv_id: str):
            """Try to scale with conversion parameter 'PAx_mV', otherwise scale = 1."""
            return int(value / parameter.get(conv_id, 1)) if value else None

        with self.connection() as con:  # commit/rollback transaction
            return insert_from_dict(
                con,
                self._table_main,
                {
                    "SetType": 1,
                    "Time": int(parametric.time * self._timebase),
                    "Status": 0,
                    "ParamID": int(parametric.param_id),
                    "PCTD": (
                        int(parametric.pctd)
                        if parametric.pctd else None
                    ),
                    "PCTA": (
                        int(parametric.pcta)
                        if parametric.pcta else None
                    ),
                    "PA0": try_convert(parametric.pa0, "PA0_mV"),
                    "PA1": try_convert(parametric.pa1, "PA1_mV"),
                    "PA2": try_convert(parametric.pa2, "PA2_mV"),
                    "PA3": try_convert(parametric.pa3, "PA3_mV"),
                    "PA4": try_convert(parametric.pa4, "PA4_mV"),
                    "PA5": try_convert(parametric.pa5, "PA5_mV"),
                    "PA6": try_convert(parametric.pa6, "PA6_mV"),
                    "PA7": try_convert(parametric.pa7, "PA7_mV"),
                },
            )



import math
from typing import Tuple

import numpy as np
from numba import njit


@njit
def hinkley(arr: np.ndarray, alpha: int = 5) -> Tuple[np.ndarray, int]:
    """
    Hinkley criterion for arrival time estimation.

    The Hinkley criterion is defined as the partial energy of the signal
    (cumulative square sum) with an applied negative trend (characterized by alpha).

    The starting value of alpha is reduced iteratively to avoid wrong picks within the
    pre-trigger part of the signal.
    Usually alpha values are chosen to be between 2 and 200 to ensure minimal delay.
    The chosen alpha value for the Hinkley criterion influences the results significantly.

    Args:
        arr: Transient signal of hit
        alpha: Divisor of the negative trend. Default: 5

    Returns:
        - Array with computed detection function
        - Index of the estimated arrival time (max value)

    Todo:
        Weak performance, if used with default parameter alpha

    References:
        - Molenda, M. (2016). Acoustic Emission monitoring of
          laboratory hydraulic fracturing experiments. Ruhr-Universität Bochum.
        - van Rijn, N. (2017).
          Investigating the Behaviour of Acoustic Emission Waves Near Cracks:
          Using the Finite Element Method. Delft University of Technology.
    """
    n = len(arr)
    result = np.zeros(n, dtype=np.float32)

    total_energy = 0.0
    for i in range(n):
        total_energy += arr[i] ** 2

    negative_trend = total_energy / (alpha * n)

    min_value = math.inf
    min_index = 0

    partial_energy = 0.0
    for i in range(n):
        partial_energy += arr[i] ** 2
        result[i] = partial_energy - (i * negative_trend)
        if result[i] < min_value:
            min_value = result[i]
            min_index = i

    return result, min_index


@njit
def aic(arr: np.ndarray) -> Tuple[np.ndarray, int]:
    """
    Akaike Information Criterion (AIC) for arrival time estimation.

    The AIC picker basically models the signal as an autoregressive (AR) process.
    A typical AE signal can be subdivided into two parts.
    The first part containing noise and the second part containing noise and the AE signal.
    Both parts of the signal contain non deterministic parts (noise) describable by a
    Gaussian distribution.

    Args:
        arr: Transient signal of hit

    Returns:
        - Array with computed detection function
        - Index of the estimated arrival time (max value)

    References:
        - Molenda, M. (2016). Acoustic Emission monitoring
          of laboratory hydraulic fracturing experiments.
          Ruhr-Universität Bochum.
        - Bai, F., Gagar, D., Foote, P., & Zhao, Y. (2017).
          Comparison of alternatives to amplitude thresholding for onset detection
          of acoustic emission signals.
          Mechanical Systems and Signal Processing, 84, 717–730.
        - van Rijn, N. (2017).
          Investigating the Behaviour of Acoustic Emission Waves Near Cracks:
          Using the Finite Element Method. Delft University of Technology.
    """
    n = len(arr)
    result = np.full(n, np.nan, dtype=np.float32)
    safety_eps = np.finfo(np.float32).tiny

    min_value = math.inf
    min_index = 0

    l_sum = 0.0
    r_sum = 0.0
    l_squaresum = 0.0
    r_squaresum = 0.0

    for i in range(n):
        r_sum += arr[i]
        r_squaresum += arr[i] ** 2

    for i in range(n - 1):
        l_sum += arr[i]
        l_squaresum += arr[i] ** 2

        r_sum -= arr[i]
        r_squaresum -= arr[i] ** 2

        l_len = i + 1
        r_len = n - i - 1

        l_variance = (1 / l_len) * l_squaresum - ((1 / l_len) * l_sum) ** 2
        r_variance = (1 / r_len) * r_squaresum - ((1 / r_len) * r_sum) ** 2

        # catch negative and very small values < safety_eps
        l_variance = max(l_variance, safety_eps)
        r_variance = max(r_variance, safety_eps)

        result[i] = (
            (i + 1) * math.log(l_variance) / math.log(10) +
            (n - i - 2) * math.log(r_variance) / math.log(10)
        )

        if result[i] < min_value:
            min_value = result[i]
            min_index = i

    return result, min_index


@njit
def energy_ratio(arr: np.ndarray, win_len: int = 100) -> Tuple[np.ndarray, int]:
    """
    Energy ratio for arrival time estimation.

    Method based on preceding and following energy collection windows.

    Args:
        arr: Transient signal of hit
        win_len: Samples of sliding windows. Default: 100

    Returns:
        - Array with computed detection function
        - Index of the estimated arrival time (max value)

    References:
        - Han, L., Wong, J., & Bancroft, J. C. (2009).
          Time picking and random noise reduction on microseismic data.
          CREWES Research Report, 21, 1–13.
    """
    n = len(arr)
    result = np.zeros(n, dtype=np.float32)
    safety_eps = np.finfo(np.float32).tiny

    max_value = -math.inf
    max_index = 0

    l_squaresum = 0.0
    r_squaresum = 0.0

    for i in range(0, win_len):
        l_squaresum += arr[i] ** 2

    for i in range(win_len, win_len + win_len):
        r_squaresum += arr[i] ** 2

    for i in range(win_len, n - win_len):
        l_squaresum += arr[i] ** 2
        r_squaresum += arr[i + win_len] ** 2
        l_squaresum -= arr[i - win_len] ** 2
        r_squaresum -= arr[i] ** 2
        result[i] = r_squaresum / (safety_eps + l_squaresum)
        if result[i] > max_value:
            max_value = result[i]
            max_index = i

    return result, max_index


@njit
def modified_energy_ratio(arr: np.ndarray, win_len: int = 100) -> Tuple[np.ndarray, int]:
    """
    Modified energy ratio method for arrival time estimation.

    The modifications improve the ability to detect the onset of a seismic
    arrival in the presence of random noise.

    Args:
        arr: Transient signal of hit
        win_len: Samples of sliding windows. Default: 100

    Returns:
        - Array with computed detection function
        - Index of the estimated arrival time (max value)

    References:
        - Han, L., Wong, J., & Bancroft, J. C. (2009).
          Time picking and random noise reduction on microseismic data.
          CREWES Research Report, 21, 1–13.
    """
    n = len(arr)
    result, _ = energy_ratio(arr, win_len)

    max_value = -math.inf
    max_index = 0

    for i in range(n):
        result[i] = (result[i] * abs(arr[i])) ** 3
        if result[i] > max_value:
            max_value = result[i]
            max_index = i

    return result, max_index



"""
IO
==

Read/write Vallen Systeme database and setup files.

Database classes
----------------

Classes to read/write pridb, tradb and trfdb database files.

.. warning:: Writing is still experimental

.. autosummary::
    :toctree: generated

    PriDatabase
    TraDatabase
    TrfDatabase

All database classes implement two different interfaces to access data:

**Standard** ``read_*``

    Read data to `pandas.DataFrame`, e.g. with `PriDatabase.read_hits`

    >>> pridb = vae.io.PriDatabase("./examples/steel_plate/sample.pridb")
    >>> df = pridb.read_hits()  # save all hits to pandas dataframe
    >>> df[["time", "channel"]]  # output columns hit and channel
                time  channel
    set_id
    10      3.992771        3
    11      3.992775        2
    12      3.992813        4
    13      3.992814        1

**Streaming** ``iread_*``

    Iterate through the data row by row.
    This is a memory-efficient solution ideal for batch processing.
    The return types are specific `typing.NamedTuple`, see :ref:`datatypes`.

    Example with `PriDatabase.iread_hits`:

    >>> pridb = vae.io.PriDatabase("./examples/steel_plate/sample.pridb")
    >>> for hit in pridb.iread_hits():
    ...     print(f"time: {hit.time:0.4f}, channel: {hit.channel}")
    ...
    time: 3.9928,   channel: 3
    time: 3.9928,   channel: 2
    time: 3.9928,   channel: 4
    time: 3.9928,   channel: 1
    >>> type(hit)
    <class 'vallenae.io.datatypes.HitRecord'>


.. _datatypes:

Data types
----------

Records of the database are represented as `typing.NamedTuple`.
Each record implements a class method ``from_sql`` to init from a SQLite
row dictionary (column name: value).

.. autosummary::
    :toctree: generated

    HitRecord
    MarkerRecord
    StatusRecord
    ParametricRecord
    TraRecord
    FeatureRecord

Compression
-----------

Transient signals in the tradb are stored as BLOBs of 16-bit ADC values --
either uncompressed or compressed (`FLAC <https://xiph.org/flac/>`_).
Following functions convert between BLOBs and arrays of voltage values.

.. autosummary::
    :toctree: generated

    decode_data_blob
    encode_data_blob
"""

# flake8: noqa

from .compression import *
from .datatypes import *
from .pridb import *
from .tradb import *
from .trfdb import *
from .types import *

__all__ = [_ for _ in dir() if not _.startswith("_")]



import math
from typing import Optional

import numpy as np
from numba import njit


@njit
def peak_amplitude(data: np.ndarray) -> float:
    """
    Compute maximum absolute amplitude.

    Args:
        data: Input array

    Returns:
        Peak amplitude of the input array
    """
    max_amplitude = 0
    for sample in data:
        abs_value = abs(sample)
        if abs_value > max_amplitude:
            max_amplitude = abs_value
    return max_amplitude


@njit
def peak_amplitude_index(data: np.ndarray) -> int:
    """
    Compute index of peak amplitude.

    Args:
        data: Input array

    Returns:
        Index of peak amplitude
    """
    max_amplitude = 0
    index_peak = 0
    for index, sample in enumerate(data):
        abs_value = abs(sample)
        if abs_value > max_amplitude:
            max_amplitude = abs_value
            index_peak = index
    return index_peak


@njit
def is_above_threshold(data: np.ndarray, threshold: float) -> bool:
    """
    Checks if absolute amplitudes are above threshold.

    Args:
        data: Input array
        threshold: Threshold amplitude

    Returns:
        True if input array is above threshold, otherwise False
    """
    for sample in data:
        if abs(sample) >= threshold:
            return True
    return False


@njit
def first_threshold_crossing(data: np.ndarray, threshold: float) -> Optional[int]:
    """
    Compute index of first threshold crossing.

    Args:
        data: Input array
        threshold: Threshold amplitude

    Returns:
        Index of first threshold crossing. None if threshold was not exceeded
    """
    for index, sample in enumerate(data):
        if abs(sample) >= threshold:
            return index
    return None


@njit
def rise_time(
    data: np.ndarray,
    threshold: float,
    samplerate: int,
    first_crossing: Optional[int] = None,
    index_peak: Optional[int] = None,
) -> float:
    """
    Compute the rise time.

    The rise time is the time between the first threshold crossing and the peak amplitude.

    Args:
        data: Input array (hit)
        threshold: Threshold amplitude (in volts)
        samplerate: Sample rate of the input array
        first_crossing: Precomputed index of first threshold crossing to save computation time
        index_peak: Precomputed index of peak amplitude to save computation time
    """
    # save some computations if pre-results are provided
    n_first_crossing = (
        first_crossing
        if first_crossing is not None
        else first_threshold_crossing(data, threshold)
    )
    n_max = (
        index_peak
        if index_peak is not None
        else peak_amplitude_index(data)
    )
    return (n_max - n_first_crossing) / samplerate


@njit
def energy(data: np.ndarray, samplerate: int) -> float:
    """
    Compute the energy of a hit.

    Energy is the integral of the squared AE-signal over time (EN 1330-9).
    The unit of energy is eu. 1 eu corresponds to 1e-14 V²s.

    Args:
        data: Input array (hit)
        samplerate: Sample rate of input array in Hz

    Returns:
        Energy of input array (hit)
    """
    agg: float = 0
    for sample in data:
        agg += sample ** 2
    return agg * 1e14 / samplerate


@njit
def signal_strength(data: np.ndarray, samplerate: int) -> float:
    """
    Compute the signal strength of a hit.

    Signal strength is the integral of the rectified AE-signal over time.
    The unit of Signal Strength is nVs (1e-9 Vs).

    Args:
        data: Input array (hit)
        samplerate: Sample rate of input array in Hz

    Returns:
        Signal strength of input array (hit)
    """
    agg: float = 0
    for sample in data:
        agg += abs(sample)
    return agg * 1e9 / samplerate


@njit
def counts(data: np.ndarray, threshold: float) -> int:
    """
    Compute the number of positive threshold crossings of a hit (counts).

    Args:
        data: Input array
        threshold: Threshold amplitude

    Returns:
        Number of positive threshold crossings
    """
    result: int = 0
    was_above_threshold: bool = False
    for sample in data:
        if sample >= threshold:
            if not was_above_threshold:
                result += 1
                was_above_threshold = True
        else:
            was_above_threshold = False
    return result


@njit
def rms(data: np.ndarray) -> float:
    """
    Compute the root mean square (RMS) of an array.

    Args:
        data: Input array

    Returns:
        RMS of the input array

    References:
        https://en.wikipedia.org/wiki/Root_mean_square
    """
    agg: float = 0
    for sample in data:
        agg += sample ** 2
    return math.sqrt(agg / len(data))
